/*
 ============================================================================
 Name	     : Engine.cpp
 Copyright   : Symbian Academy, 2007
 Description : Game engine implementation
 ============================================================================
 */

#include "Engine.h"

CEngine::CEngine(CConsoleBase* aConsole) :
	iConsole(aConsole)
	{
	// No implementation required
	}

CEngine::~CEngine()
	{
	delete iStreetPos;
	iPlayerGraphics.Close ();
	iSpaceLine.Close ();
	}

CEngine* CEngine::NewLC(CConsoleBase* aConsole)
	{
	CEngine* self = new (ELeave)CEngine(aConsole);
	CleanupStack::PushL (self);
	self->ConstructL ();
	return self;
	}

CEngine* CEngine::NewL(CConsoleBase* aConsole)
	{
	CEngine* self=CEngine::NewLC (aConsole);
	CleanupStack::Pop (); // self;
	return self;
	}

void CEngine::ConstructL()
	{

	}

void CEngine::SetUpdateTimer(CUpdateTimer* aUpdateTimer)
	{
	iUpdateTimer = aUpdateTimer;
	}

void CEngine::GameOver(TInt /*aReason*/)
	{
	// Write "Game Over" centered on the screen
	iConsole->SetCursorPosAbs (TPoint ( (iWidth - KGameOver().Length()) / 2, iHeight / 2));
	iConsole->Write (KGameOver);
	iConsole->SetCursorPosAbs (TPoint ( 0, iHeight - 4));
	// Edit 23: Stop the active scheduler so that the program continues
	//          at the position where the scheduler was started.
	//          -> the game leaves MainL() -> E32Main() and then exits.
	
	}

void CEngine::SetParametersL(TInt aEnergy, TInt aHeight, TInt aWidth,
		TInt aStreetWidth, TInt aMaxDuration, TInt aPlayerWidth)
	{
	// Statistics
	iEnergy = aEnergy;
	iLevel = 1;
	iScore = 0;
	
	// Parameters of the game field
	iHeight = aHeight;
	iWidth = aWidth;
	iStreetWidth = aStreetWidth;

	// Player settings
	iPlayerWidth = aPlayerWidth;
	iPlayerPos = (iWidth - iPlayerWidth) / 2;
	iLastPlayerPos = iPlayerPos;
	iPlayerGraphics.CreateMaxL (iPlayerWidth);
	iPlayerGraphics.Fill ('O');

	// Empty line used for clearing parts of the game field
	iSpaceLine.CreateMaxL (iWidth);
	iSpaceLine.Fill (' ');

	// Street
	// How long to continue in one direction before calculating again
	if ( aMaxDuration <= 3)
		User::Leave (KErrArgument);
	iMaxDuration = aMaxDuration;
	// Build the street
	iStreetPos = new (ELeave) TInt[iHeight];
	// Set the first position to the center
	iStreetPos[0] = (iWidth - iStreetWidth) / 2;
	iCurArrayPos = 0;
	iDurationLeft = 0;

	// Call the update function once for every line, so that it
	// builds the whole screen.
	for (TInt i = 0; i < iHeight - 1; i++)
		{
		NewStreetValue ();
		}
	}

void CEngine::NewStreetValue()
	{
	// Save the previous position of the array whill will be
	// kept for the new line, increased (right) or decreased (left)
	const TInt prevArrayPos = iCurArrayPos;

	// Circular array!
	iCurArrayPos ++;
	if ( iCurArrayPos >= iHeight)
		iCurArrayPos = 0;

	// Check if the duration of the previous movement is over.
	if ( iDurationLeft <= 0)
		{
		// If yes: think of a new direction and duration
		TInt maxDuration;
		do
			{
			// New direction
			iCurDirection = GetRandomNumber (-1, 1);
			// Check how long the street can move in this direction
			maxDuration = GetMaxDurationForDirection (iStreetPos[prevArrayPos],
					iCurDirection);
			// Check if we're not already at the border - if yes, do the
			// same calculation again. Could be made a bit more efficient.
			}
		while (maxDuration <= 1);
		// Save the new duration for this movement, a random number between 1
		// and the maximum duration calculated above.
		iDurationLeft = GetRandomNumber (1, maxDuration);
		}

	// Assign a new value to the next street position
	iStreetPos[iCurArrayPos] = iStreetPos[prevArrayPos] + iCurDirection;
	iDurationLeft --;
	}

TInt CEngine::GetRandomNumber(const TInt aLow, const TInt aHigh)
	{
	return Math::Random ()% (aHigh - aLow + 1) + aLow;
	}

TInt CEngine::GetMaxDurationForDirection(const TInt aCurX, const TInt aDirection)
	{
	// For straight movement
	TInt maxDuration = iMaxDuration;
	switch (aDirection)
		{
		case -1: // Left
			maxDuration = aCurX - 1;
			break;
		case 1: // Right - don't forget to take the street width into account!
			maxDuration = iWidth - iStreetWidth - aCurX - 1;
			break;
		}
	return maxDuration;
	}

void CEngine::KeyPressed(TKeyCode aKey)
	{
	switch (aKey)
		{
		case EKeyLeftArrow: // Move the player left by one char
			iPlayerPos --;
			if ( iPlayerPos < 1)
				iPlayerPos = 1;
			break;
		case EKeyRightArrow: // Move the player right by one char
			iPlayerPos ++;
			if ( iPlayerPos > iWidth - iPlayerWidth - 1)
				iPlayerPos = iWidth - iPlayerWidth - 1;
			break;
		case ' ': // Escape, Space, q, x: Exit
		case 'q':
		case 'x':
		case EKeyEscape:
			GameOver (0);
			break;
		}
	}

void CEngine::UpdateWorldL()
	{
	// Update the street
	NewStreetValue ();

	// Calculate the next array start position.
	TInt arrayPos = iCurArrayPos + 1;
	if ( arrayPos >= iHeight)
		arrayPos = 0;

	// Score and speed update
	iScore ++;
	if ( iScore % 100 == 0)
		{
		if ( iUpdateTimer)
			{
			iUpdateTimer->IncreaseSpeed (1000);
			iLevel++;
			if (iLevel % 5 == 0)
				{
				iStreetWidth --;
				// Clear the screen to prevent artifacts from a non-deleted
				// street (it's now instantly smaller, so the " V " descriptor 
				// won't remove everything.
				iConsole->ClearScreen();
				}
			}
		}
	// Collision detection
	const TInt curStreetPos = iStreetPos[arrayPos];
	if ( iPlayerPos < curStreetPos || (iPlayerPos + iPlayerWidth) > (curStreetPos + iStreetWidth + 1))
		{
		iEnergy --;
		}

	// Player
	// Clear previous stuff if the player moved
	// Otherwise, don't redraw the player to prevent flickering
	const TInt playerMovement = iPlayerPos - iLastPlayerPos;
	if ( playerMovement > 0)
		{
		// Movement to the right
		TPtrC removeTxt(iSpaceLine.Ptr (), playerMovement);
		iConsole->SetCursorPosAbs (TPoint (iLastPlayerPos, iHeight - 1));
		iConsole->Write (removeTxt);
		}
	else
		if ( playerMovement < 0)
			{
			// Movement to the left
			TPtrC removeTxt(iSpaceLine.Ptr (), -playerMovement);
			iConsole->SetCursorPosAbs (TPoint (iPlayerPos + iPlayerWidth,
					iHeight - 1));
			iConsole->Write (removeTxt);
			}
	iLastPlayerPos = iPlayerPos;

	// Draw the street
	TInt y;
	for (y = iHeight - 1; y > 0; y--)
		{
		// Left border
		iConsole->SetCursorPosAbs (TPoint (iStreetPos[arrayPos] - 1, y));
		iConsole->Write (KStreetBorder);
		// Right border
		iConsole->SetCursorPosAbs (TPoint (
				iStreetPos[arrayPos] - 1 + iStreetWidth, y));
		iConsole->Write (KStreetBorder);
		// Update current circular array pos.
		arrayPos ++;
		if ( arrayPos >= iHeight)
			arrayPos = 0;
		}

	// Draw the player to the screen
	iConsole->SetCursorPosAbs (TPoint (iPlayerPos, iHeight - 1));
	iConsole->Write (iPlayerGraphics);

	// Draw the status message
	iConsole->SetCursorPosAbs (TPoint (0, 0));
	iConsole->Printf (KStatus, iScore, iLevel, iEnergy);

	// Check if the energy is gone. Do this after drawing to make sure that
	// the whole screen is updated before exiting the game.
	if ( iEnergy == 0)
		GameOver (1);
	}
